/*************************************************************************\
* Copyright (c) 2009 UChicago Argonne LLC, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE is distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/* dbBase.h,v 1.12.2.8 2009/03/19 23:00:44 anj Exp
 *
 *      Current Author:         Nick Rees
 *      Date:                   01-Apr-2009
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dbStaticLib.h"
#include "dbAccess.h"
#include "drvSup.h"
#include "epicsExport.h"


#define CVT_DBADDR_OK         (0x1 << 1 )
#define GET_ARRAY_INFO_OK     (0x1 << 2 )
#define PUT_ARRAY_INFO_OK     (0x1 << 3 )
#define GET_UNITS_OK          (0x1 << 4 )
#define GET_PRECISION_OK      (0x1 << 5 )
#define GET_ENUM_STR_OK       (0x1 << 6 )
#define GET_ENUM_STRS_OK      (0x1 << 7 )
#define PUT_ENUM_STR_OK       (0x1 << 8 )
#define GET_GRAPHIC_DOUBLE_OK (0x1 << 9 )
#define GET_CONTROL_DOUBLE_OK (0x1 <<10 )
#define GET_ALARM_DOUBLE_OK   (0x1 <<11 )

#define fieldPtr( paddr, pfldDes, default ) ((pfldDes) == NULL ? &(default) : fldSupPtr( (paddr), (pfldDes) ))

/* static long report(FILE fp, struct dbFldDes * dbFldDes); */
/* long (*init)(void) */
static void * init_field( DBBASE *pdbbase, struct dbFldDes *);
/* static long init_record( struct dbAddr *, int pass ); */
/* static long special( const struct dbAddr *paddr, int after ); */
static long cvt_dbaddr( struct dbAddr * paddr );
static long get_array_info( const struct dbAddr *paddr, long *no_elements, long *offset);
static long put_array_info( const struct dbAddr *paddr, long nRequest );
static long get_units(const struct dbAddr *paddr, char *units);
static long get_precision(const struct dbAddr *paddr, long *precision);
/* static long get_enum_str(const struct dbAddr *paddr, char * pbuffer); */
/* static long get_enum_strs(const struct dbAddr *paddr, struct dbr_enumStrs *p); */
/* static long put_enum_str(const struct dbAddr *paddr, const char * pbuffer); */
static long get_graphic_double(const struct dbAddr *paddr, struct dbr_grDouble *pgd);
static long get_control_double(const struct dbAddr *paddr, struct dbr_ctrlDouble *pcd);
static long get_alarm_double(const struct dbAddr *paddr,struct dbr_alDouble *p);

fldSup_SET fldSupDefault_fset = {
    16,
    NULL,               /* long (*report)(FILE fp, struct dbFldDes * dbFldDes);                                  */
    NULL,               /* long (*init)(void);                                                                   */
    init_field,         /* void * (*init_field)( DBBASE *pdbbase, struct dbFldDes *);                            */
    NULL,               /* long (*init_record)( struct dbAddr *, int pass );                                     */
    NULL,               /* long (*special)( const struct dbAddr *paddr, int after );                             */
    cvt_dbaddr,         /* long (*cvt_dbaddr)( struct dbAddr * paddr );                                          */
    get_array_info,     /* long (*get_array_info)( const struct dbAddr *paddr, long *no_elements, long *offset); */
    put_array_info,     /* long (*put_array_info)( const struct dbAddr *paddr, long nRequest );                  */
    get_units,          /* long (*get_units)(const struct dbAddr *paddr, char *units);                           */
    get_precision,      /* long (*get_precision)(const struct dbAddr *paddr, long *precision);                   */
    NULL,               /* long (*get_enum_str)(const struct dbAddr *paddr, char * pbuffer);                     */
    NULL,               /* long (*get_enum_strs)(const struct dbAddr *paddr, struct dbr_enumStrs *p);            */
    NULL,               /* long (*put_enum_str)(const struct dbAddr *paddr, const char * pbuffer);               */
    get_graphic_double, /* long (*get_graphic_double)(const struct dbAddr *paddr, struct dbr_grDouble *pgd);     */
    get_control_double, /* long (*get_control_double)(const struct dbAddr *paddr, struct dbr_ctrlDouble *pcd);   */
    get_alarm_double    /* long (*get_alarm_double)(const struct dbAddr *paddr,struct dbr_alDouble *p);          */
};

epicsExportAddress(fldset,fldSupDefault_fset);

typedef struct {
    unsigned long flags;
    dbFldDes * bptrFld;
    epicsEnum16 type;
    dbFldDes * typeFld;
    epicsUInt32 nelm;
    dbFldDes * nelmFld;
    epicsUInt32 nord;
    dbFldDes * nordFld;
    char * units;
    dbFldDes * unitsFld;
    epicsUInt32 prec;
    dbFldDes * precFld;
    epicsFloat64 hopr;
    dbFldDes * hoprFld;
    epicsFloat64 lopr;
    dbFldDes * loprFld;
    epicsFloat64 drvh;
    dbFldDes * drvhFld;
    epicsFloat64 drvl;
    dbFldDes * drvlFld;
    epicsFloat64 hihi;
    dbFldDes * hihiFld;
    epicsFloat64 high;
    dbFldDes * highFld;
    epicsFloat64 low;
    dbFldDes * lowFld;
    epicsFloat64 lolo;
    dbFldDes * loloFld;
} dbAddrPvt;


static int initUInt32( DBBASE *pdbbase, dbFldDes * pfldDes, char * name, dbFldDes ** ppField, epicsUInt32 * value )
{
    char * string;
    fldSupParamId pParam = pfldDes->fldSupParams;
    int ok = 0;

   if ( ( string = fldSupParamGet( pParam, name )) != 0 ) {
        ok = ( sscanf( string, "%u", value ) == 1 );

        if  ( !ok ) {
            *ppField = fldSupFindFld( pdbbase, pfldDes, string );
            ok = ( *ppField != NULL && (*ppField)->field_type == DBF_ULONG );

            if ( !ok ) {
                errlogPrintf("init_long: %s not an unsigned long in record type %s, field %s",
                             name, pfldDes->pdbRecordType->name, pfldDes->name );
                *ppField = NULL;
            }
        }
    }

   return ok;
}

static int initFloat64( DBBASE *pdbbase, dbFldDes * pfldDes, char * name, dbFldDes ** ppField, epicsFloat64 * value )
{
    char * string;
    fldSupParamId pParam = pfldDes->fldSupParams;
    int ok = 0;

   if ( ( string = fldSupParamGet( pParam, name )) != 0 ) {
        ok = ( sscanf( string, "%lf", value ) == 1 );

        if  ( !ok ) {
            *ppField = fldSupFindFld( pdbbase, pfldDes, string );
            ok = ( *ppField != NULL && (*ppField)->field_type == DBF_DOUBLE );

            if ( !ok ) {
                errlogPrintf("init_long: %s not a double in record type %s, field %s",
                             name, pfldDes->pdbRecordType->name, pfldDes->name );
                *ppField = NULL;
            }
        }
    }

   return ok;
}


static void * init_field( DBBASE *pdbbase, dbFldDes * pfldDes)
{
    char * value;
    int i;
    fldSupParamId pParam = pfldDes->fldSupParams;
    dbAddrPvt * pPvt = calloc( 1, sizeof(dbAddrPvt) );
    int bptrOK=0, nelmOK=0, nordOK=0, typeOK=0, precOK=0;
    int hoprOK=0, loprOK=0, drvhOK=0, drvlOK=0, unitsOK=0;
    int hihiOK=0, highOK=0, lowOK=0,  loloOK=0;

    if ( pParam == NULL ) return NULL;

    if ( (value = fldSupParamGet( pParam, "bptr" )) != NULL ) {
        
        if ((pPvt->bptrFld = fldSupFindFld( pdbbase, pfldDes, value )) != NULL ) {
            bptrOK = (pPvt->bptrFld->field_type == DBF_NOACCESS);

            if (!bptrOK) {
                errlogPrintf("dbaddr_plugin_init: BPTR not a pointer in record type %s, field %s",
                             pfldDes->pdbRecordType->name, pfldDes->name );
                pPvt->bptrFld = NULL;
            }
        }
    }

    if ( (value = fldSupParamGet( pParam, "type" )) != 0 ) {
        for(i=0; i<DBF_NTYPES && ! typeOK; i++) {
            typeOK = (strcmp(value,pamapdbfType[i].strvalue)==0);
            if (typeOK) pPvt->type  = i;
        }
 
        if ( !typeOK ) {
            pPvt->typeFld = fldSupFindFld( pdbbase, pfldDes, value );
            typeOK = ( pPvt->typeFld != NULL && pPvt->typeFld->field_type == DBF_MENU);

            if ( ! typeOK ) {
                errlogPrintf("dbaddr_plugin_init: type %s not a short in record type %s, field %s",
                             value, pfldDes->pdbRecordType->name, pfldDes->name );
                pPvt->typeFld = NULL;
            }
            else
            {
                printf( "dbaddr_plugin_init: setting record type %s, field %s type to %s",
                        pfldDes->pdbRecordType->name, pfldDes->name, value );
            }
        }
    }

    value = fldSupParamGet( pParam, "units" );
    unitsOK = ( value != NULL );

    if (unitsOK) {        
        if ((pPvt->unitsFld = fldSupFindFld( pdbbase, pfldDes, value )) == NULL ) {
            pPvt->units = value;
        }
    }

    nelmOK = initUInt32( pdbbase, pfldDes, "nelm", &(pPvt->nelmFld), &(pPvt->nelm) );
    nordOK = initUInt32( pdbbase, pfldDes, "nord", &(pPvt->nordFld), &(pPvt->nord) );
    precOK = initUInt32( pdbbase, pfldDes, "prec", &(pPvt->precFld), &(pPvt->prec) );
    hoprOK = initFloat64( pdbbase, pfldDes, "hopr", &(pPvt->hoprFld), &(pPvt->hopr) );
    loprOK = initFloat64( pdbbase, pfldDes, "lopr", &(pPvt->loprFld), &(pPvt->lopr) );
    drvhOK = initFloat64( pdbbase, pfldDes, "drvh", &(pPvt->drvhFld), &(pPvt->drvh) );
    drvlOK = initFloat64( pdbbase, pfldDes, "drvl", &(pPvt->drvlFld), &(pPvt->drvl) );
    hihiOK = initFloat64( pdbbase, pfldDes, "hihi", &(pPvt->hihiFld), &(pPvt->hihi) );
    highOK = initFloat64( pdbbase, pfldDes, "high", &(pPvt->highFld), &(pPvt->high) );
    lowOK  = initFloat64( pdbbase, pfldDes, "low", &(pPvt->lowFld), &(pPvt->low) );
    loloOK = initFloat64( pdbbase, pfldDes, "lolo", &(pPvt->loloFld), &(pPvt->lolo) );

    if ( bptrOK && nelmOK && typeOK )          pPvt->flags |= CVT_DBADDR_OK;
    if ( nordOK )                              pPvt->flags |= GET_ARRAY_INFO_OK;
    if ( nordOK && nelmOK )                    pPvt->flags |= PUT_ARRAY_INFO_OK;
    if ( precOK )                              pPvt->flags |= GET_PRECISION_OK;
    if ( hoprOK && loprOK )                    pPvt->flags |= GET_GRAPHIC_DOUBLE_OK;
    if ( drvhOK && drvlOK )                    pPvt->flags |= GET_CONTROL_DOUBLE_OK;
    if ( unitsOK )                             pPvt->flags |= GET_UNITS_OK;
    if ( hihiOK && highOK && lowOK && loloOK ) pPvt->flags |= GET_ALARM_DOUBLE_OK;

    pfldDes->fldSupPvt = pPvt;
    return pPvt;
}

static long cvt_dbaddr( DBADDR *paddr)
{
    dbAddrPvt * pPvt = fldSupPvt( paddr );
    long status = S_db_noRSET;

    if ( pPvt->flags & CVT_DBADDR_OK )
    {
        paddr->pfield = * (void ** ) fldSupPtr( paddr, pPvt->bptrFld );
        paddr->no_elements = * (epicsUInt32 *) fieldPtr( paddr, pPvt->nelmFld, pPvt->nelm );
        paddr->field_type = * (epicsEnum16 *) fieldPtr( paddr, pPvt->typeFld, pPvt->type );
        paddr->field_size = dbValueSize(paddr->field_type);
        paddr->dbr_field_type = paddr->field_type;
        status = 0;

        printf( "cvt_dbaddr: returning NELM of %ld, TYPE of %d, SIZE of %d\n",
                paddr->no_elements, (int) paddr->field_type, (int) paddr->field_size );
    }

    return status;
}


static long get_array_info( const struct dbAddr *paddr, long *no_elements, long *offset)
{
    dbAddrPvt * pPvt = fldSupPvt( paddr );
    long status = S_db_noRSET;

    if ( pPvt->flags & GET_ARRAY_INFO_OK )
    {
        epicsUInt32 * pnord = (epicsUInt32 *) fieldPtr( paddr, pPvt->nordFld, pPvt->nord );

        *no_elements = *pnord;
        *offset = 0;
        status = 0;

        printf( "get_array_info: Returning NORD of %ud\n", *pnord );
    }

    return status;
}

static long put_array_info( const struct dbAddr *paddr, long nRequest )
{
    dbAddrPvt * pPvt = fldSupPvt( paddr );
    long status = S_db_noRSET;

    if ( pPvt->flags & PUT_ARRAY_INFO_OK )
    {
        epicsUInt32 * pnord = (epicsUInt32 *) fieldPtr( paddr, pPvt->nordFld, pPvt->nord );
        epicsUInt32 * pnelm = (epicsUInt32 *) fieldPtr( paddr, pPvt->nelmFld, pPvt->nelm );

        *pnord = nRequest;
        if (*pnord > *pnelm)
            *pnord = *pnelm;
        status = 0;

        printf( "put_array_info: Setting NORD to %ud\n", *pnord );
    }

    return status;
}

static long get_units(const DBADDR *paddr, char *units)
{
    dbAddrPvt * pPvt = fldSupPvt( paddr );
    long status = S_db_noRSET;

    if ( pPvt->flags & GET_UNITS_OK )
    {
        char * unitsVal = (char *) fieldPtr( paddr, pPvt->unitsFld, *(pPvt->units) );
        strncpy(units, unitsVal, DB_UNITS_SIZE);
        status = 0;

        printf( "put_array_info: Setting UNITS to %s\n", units );
    }

    return status;
}

static long get_precision(const DBADDR *paddr, long *precision)
{
    dbAddrPvt * pPvt = fldSupPvt( paddr );
    long status = S_db_noRSET;

    if ( pPvt->flags & GET_PRECISION_OK )
    {
        epicsUInt32 * pprec = (epicsUInt32 *) fieldPtr( paddr, pPvt->precFld, pPvt->prec );
        *precision = *pprec;
        status = 0;

        printf( "get_precision: Setting precision to %ud\n", *pprec );
    }

    return status;
}


static long get_graphic_double(const DBADDR *paddr, struct dbr_grDouble *pgd)
{
    dbAddrPvt * pPvt = fldSupPvt( paddr );
    long status = S_db_noRSET;

    if ( pPvt->flags & GET_GRAPHIC_DOUBLE_OK )
    {
        epicsFloat64 * phopr = (epicsFloat64 *) fieldPtr( paddr, pPvt->hoprFld, pPvt->hopr );
        epicsFloat64 * plopr = (epicsFloat64 *) fieldPtr( paddr, pPvt->loprFld, pPvt->lopr );

        pgd->upper_disp_limit = *phopr;
        pgd->lower_disp_limit = *plopr;
        status = 0;

        printf( "get_graphic_double: Returning hopr = %f, lopr = %f\n",
                pgd->upper_disp_limit, pgd->lower_disp_limit );
    }

    return status;
}

static long get_control_double(const DBADDR *paddr, struct dbr_ctrlDouble *pcd)
{
    dbAddrPvt * pPvt = fldSupPvt( paddr );
    long status = S_db_noRSET;

    if ( pPvt->flags & GET_GRAPHIC_DOUBLE_OK )
    {
        epicsFloat64 * pdrvh = (epicsFloat64 *) fieldPtr( paddr, pPvt->drvhFld, pPvt->drvh );
        epicsFloat64 * pdrvl = (epicsFloat64 *) fieldPtr( paddr, pPvt->drvlFld, pPvt->drvl );

        pcd->upper_ctrl_limit = *pdrvh;
        pcd->lower_ctrl_limit = *pdrvl;
        status = 0;

        printf( "get_control_double: Returning drvh = %f, drvl = %f\n",
                pcd->upper_ctrl_limit, pcd->lower_ctrl_limit );
    }

    return status;
}

static long get_alarm_double(const DBADDR *paddr, struct dbr_alDouble *pad)
{
    dbAddrPvt * pPvt = fldSupPvt( paddr );
    long status = S_db_noRSET;

    if ( pPvt->flags & GET_ALARM_DOUBLE_OK )
    {
	pad->upper_alarm_limit   = * (epicsFloat64 *) fieldPtr( paddr, pPvt->hihiFld, pPvt->hihi );
	pad->upper_warning_limit = * (epicsFloat64 *) fieldPtr( paddr, pPvt->highFld, pPvt->high );
	pad->lower_warning_limit = * (epicsFloat64 *) fieldPtr( paddr, pPvt->lowFld,  pPvt->low  );
	pad->lower_alarm_limit   = * (epicsFloat64 *) fieldPtr( paddr, pPvt->loloFld, pPvt->lolo );
        status = 0;

        printf( "get_alarm_double: Returning hihi = %f, high = %f, low = %f, lolo = %f\n",
                pad->upper_alarm_limit,   pad->upper_warning_limit,
                pad->lower_warning_limit, pad->lower_alarm_limit    );
    }

    return status;
}
